;;;; fp-utils.scm
;;;; Kon Lovett, May '17
;;;; Kon Lovett, Sep '17

;;;; Issues
;;;;

(module fp-utils

(;export
  ;
  fprandom
  ;
  fpzero? fppositive? fpcardinal? fpnegative? fpnon-positive?
  ;
  fpeven? fpodd?
  ;
  fpclosed-right? fpclosed? fpclosed-left? fpclosedr? fpclosedl?
  ;
  fpadd1 fpsub1
  ;
  fpmodulo
  fpquotient fpremainder
  ;
  fpfraction
  ;
  fptruncate/precision fpround/precision fpceiling/precision fpfloor/precision
  ;
  fp~= fp~<= fp~>=
  ;
  fpsqr fpcub
  ;
  fpdegree->radian fpradian->degree
  ;
  fpdistance fpdistance*
  ;
  fpmax-and-min
  ;
  fpprecision-factor)

(import scheme)

(import chicken foreign extras)

(declare
  (bound-to-procedure
    ##sys#flonum-fraction
    ##sys#check-inexact) )

;;;

(define C_fmod
  (foreign-lambda double "fmod" double double))

(define C_remainder
  (foreign-lambda double "remainder" double double))

(: *fpeven? (float --> boolean))
(define (*fpeven? n)
  (fp= 0.0 (##sys#flonum-fraction (fp/ n 2.0))) )

;;;

;;

(: fprandom (#!optional (or float fixnum) --> float))
(define (fprandom #!optional (x most-positive-fixnum))
  ;
  (define log10
    (let ((ln10 (log 10)))
      (lambda (n)
        (fp/ (log n) ln10) ) ) )
  ;
  (let ((x
          (cond
            ((flonum? x)
              (inexact->exact (expt 10 (abs (round (log10 x))))) )
            (else ;(fixnum? x)
              x ) ) ) )
    (fp/ 1.0 (exact->inexact (random x))) ) )

;;

(: fpzero? (float --> boolean))
(define (fpzero? n)
  (fp= 0.0 n) )

(: fppositive? (float --> boolean))
(define (fppositive? n)
  (fp< 0.0 n) )

(: fpcardinal? (float --> boolean))
(define (fpcardinal? n)
  (fp<= 0.0 n) )

(: fpnegative? (float --> boolean))
(define (fpnegative? n)
  (fp> 0.0 n) )

(: fpnon-positive? (float --> boolean))
(define (fpnon-positive? n)
  (fp>= 0.0 n) )

;;

(: fpeven? (float --> boolean))
(define (fpeven? n)
  (and
    (fpinteger? n)
    (*fpeven? n)) )

(: fpodd? (float --> boolean))
(define (fpodd? n)
  (and
    (fpinteger? n)
    (not (*fpeven? n))) )

;;

(: fpclosed-right? (float float float --> boolean))
(define (fpclosed-right? l x h)
  (and (fp< l x) (fp<= x h)) )

(: fpclosed? (float float float --> boolean))
(define (fpclosed? l x h)
  (and (fp<= l x) (fp<= x h)) )

(: fpclosed-left? (float float float --> boolean))
(define (fpclosed-left? l x h)
  (and (fp<= l x) (fp< x h)) )

(: fpclosedr? (float float float --> boolean))
(define fpclosedr? fpclosed-right?)

(: fpclosedl? (float float float --> boolean))
(define fpclosedl? fpclosed-left?)

;;

(: fpadd1 (float --> float))
(define (fpadd1 n)
  (fp+ n 1.0) )

(: fpsub1 (float --> float))
(define (fpsub1 n)
  (fp- n 1.0) )

;;

(: fpmodulo (float float --> float))
(define (fpmodulo x y)
  (##sys#check-inexact x 'fpmodulo)
  (##sys#check-inexact y 'fpmodulo)
  (fptruncate (C_fmod x y)) )

(: fpquotient (float float --> float))
(define (fpquotient x y)
  (fptruncate (fp/ x y)) )

(: fpremainder (float float --> float))
(define (fpremainder x y)
  (##sys#check-inexact x 'fpremainder)
  (##sys#check-inexact y 'fpremainder)
  (fptruncate (C_remainder x y)) )

;;

(: fpfraction (float --> float))
(define (fpfraction n)
	(##sys#flonum-fraction n) )

;;;

;;

(: fp~= (float float #!optional float --> boolean))
(define (fp~= x y #!optional (eps flonum-epsilon))
  (let ((diff (fp- x y)))
    (or
      (fpzero? diff)
      (fp<= (fpabs diff) eps) ) ) )

(: fp~<= (float float #!optional float --> boolean))
(define (fp~<= x y #!optional (eps flonum-epsilon))
  (or
    (fp< x y)
    (fp~= x y eps) ) )

(: fp~>= (float float #!optional float --> boolean))
(define (fp~>= x y #!optional (eps flonum-epsilon))
  (or
    (fp> x y)
    (fp~= x y eps) ) )

;;;

;;

(: fpsqr (float --> float))
(define (fpsqr n)
  (fp* n n) )

(: fpcub (float --> float))
(define (fpcub n)
  (fp* n (fp* n n)) )

;;;

(define-constant PRECISION-DEFAULT 4.0)

(define-syntax make-unary/precision
  (syntax-rules ()
    ((_ ?op)
      (lambda (n #!optional (p PRECISION-DEFAULT))
        (if (fpzero? p)
          (?op n)
          (let ((precfact (fpprecision-factor p)))
            (fp/ (?op (fp* n precfact)) precfact) ) ) ) ) ) )

;;

(: fptruncate/precision (float #!optional float --> float))
(define fptruncate/precision (make-unary/precision fptruncate))

(: fpround/precision (float #!optional float --> float))
(define fpround/precision (make-unary/precision fpround))

(: fpceiling/precision (float #!optional float --> float))
(define fpceiling/precision (make-unary/precision fpceiling))

(: fpfloor/precision (float #!optional float --> float))
(define fpfloor/precision (make-unary/precision fpfloor))

;;

(define-constant DEGREE 0.0174532925199432957692369076848861271344) ;pi/180

(: fpdegree->radian (float --> float))
(define (fpdegree->radian deg)
  (fp* deg DEGREE) )

(: fpradian->degree (float --> float))
(define (fpradian->degree rad)
  (fp/ rad DEGREE) )

;;

(: fpdistance (float float float float --> float))
(define (fpdistance x1 y1 x2 y2)
  (fpsqrt (fpdistance* x1 y1 x2 y2)) )

(: fpdistance* (float float float float --> float))
(define (fpdistance* x1 y1 x2 y2)
  (fp+ (fpsqr (fp- x1 x2)) (fpsqr (fp- y1 y2))) )

;;

(: fpmax-and-min (float #!rest float --> float float))
(define (fpmax-and-min fp . fps)
  (let loop ((fps fps) (mx fp) (mn fp))
    (if (null? fps)
      (values mx mn)
      (let ((cur (car fps)))
        (loop (cdr fps) (fpmax mx cur) (fpmin mn cur)) ) ) ) )

;;

(: fpprecision-factor ((or float fixnum) #!optional float --> float))
(define (fpprecision-factor p #!optional (base 10.0))
  (fpexpt base (exact->inexact p)) )

) ;fp-utils
